@@ -1,4 +1,3 @@
-#include <linux/nl80211.h>
 #include <net/if.h>
 #include <errno.h>
 #include <string.h>
@@ -9,10 +8,11 @@
 #include <netlink/msg.h>
 #include <netlink/attr.h>
 
+#include "nl80211.h"
 #include "iw.h"
 
 static char *mntr_flags[NL80211_MNTR_FLAG_MAX + 1] = {
-	NULL,
+	"none",
 	"fcsfail",
 	"plcpfail",
 	"control",
@@ -20,6 +20,56 @@ static char *mntr_flags[NL80211_MNTR_FLAG_MAX + 1] = {
 	"cook",
 };
 
+static int parse_mntr_flags(int *_argc, char ***_argv,
+			    struct nl_msg *msg)
+{
+	struct nl_msg *flags;
+	int err = -ENOBUFS;
+	enum nl80211_mntr_flags flag;
+	int argc = *_argc;
+	char **argv = *_argv;
+
+	flags = nlmsg_alloc();
+	if (!flags)
+		return -ENOMEM;
+
+	while (argc) {
+		int ok = 0;
+		for (flag = __NL80211_MNTR_FLAG_INVALID;
+		     flag <= NL80211_MNTR_FLAG_MAX; flag++) {
+			if (strcmp(*argv, mntr_flags[flag]) == 0) {
+				ok = 1;
+				/*
+				 * This shouldn't be adding "flag" if that is
+				 * zero, but due to a problem in the kernel's
+				 * nl80211 code (using NLA_NESTED policy) it
+				 * will reject an empty nested attribute but
+				 * not one that contains an invalid attribute
+				 */
+				NLA_PUT_FLAG(flags, flag);
+				break;
+			}
+		}
+		if (!ok) {
+			err = -EINVAL;
+			goto out;
+		}
+		argc--;
+		argv++;
+	}
+
+	nla_put_nested(msg, NL80211_ATTR_MNTR_FLAGS, flags);
+	err = 0;
+ nla_put_failure:
+ out:
+	nlmsg_free(flags);
+
+	*_argc = argc;
+	*_argv = argv;
+
+	return err;
+}
+
 /* return 0 if not found, 1 if ok, -1 on error */
 static int get_if_type(int *argc, char ***argv, enum nl80211_iftype *type)
 {
@@ -42,6 +92,14 @@ static int get_if_type(int *argc, char ***argv, enum nl80211_iftype *type)
 	} else if (strcmp(tpstr, "monitor") == 0) {
 		*type = NL80211_IFTYPE_MONITOR;
 		return 1;
+	} else if (strcmp(tpstr, "master") == 0) {
+		*type = NL80211_IFTYPE_UNSPECIFIED;
+		fprintf(stderr, "See http://wireless.kernel.org/RTFM-AP.\n");
+		return 2;
+	} else if (strcmp(tpstr, "ap") == 0) {
+		*type = NL80211_IFTYPE_UNSPECIFIED;
+		fprintf(stderr, "See http://wireless.kernel.org/RTFM-AP.\n");
+		return 2;
 	} else if (strcmp(tpstr, "__ap") == 0) {
 		*type = NL80211_IFTYPE_AP;
 		return 1;
@@ -65,7 +123,8 @@ static int get_if_type(int *argc, char ***argv, enum nl80211_iftype *type)
 	return -1;
 }
 
-static int handle_interface_add(struct nl_cb *cb,
+static int handle_interface_add(struct nl80211_state *state,
+				struct nl_cb *cb,
 				struct nl_msg *msg,
 				int argc, char **argv)
 {
@@ -82,20 +141,29 @@ static int handle_interface_add(struct nl_cb *cb,
 	argv++;
 
 	tpset = get_if_type(&argc, &argv, &type);
-	if (tpset <= 0)
-		return 1;
+	if (tpset != 1)
+		return tpset;
 
 	if (argc) {
-		if (strcmp(argv[0], "mesh_id") != 0)
-			return 1;
-		argc--;
-		argv++;
+		if (strcmp(argv[0], "mesh_id") == 0) {
+			argc--;
+			argv++;
 
-		if (!argc)
+			if (!argc)
+				return 1;
+			mesh_id = argv[0];
+			argc--;
+			argv++;
+		} else if (strcmp(argv[0], "flags") == 0) {
+			argc--;
+			argv++;
+			if (parse_mntr_flags(&argc, &argv, msg)) {
+				fprintf(stderr, "flags error\n");
+				return 2;
+			}
+		} else {
 			return 1;
-		mesh_id = argv[0];
-		argc--;
-		argv++;
+		}
 	}
 
 	if (argc)
@@ -111,12 +179,13 @@ static int handle_interface_add(struct nl_cb *cb,
  nla_put_failure:
 	return -ENOBUFS;
 }
-COMMAND(interface, add, "<name> type <type> [mesh_id <meshid>]",
+COMMAND(interface, add, "<name> type <type> [mesh_id <meshid>] [flags ...]",
 	NL80211_CMD_NEW_INTERFACE, 0, CIB_PHY, handle_interface_add);
-COMMAND(interface, add, "<name> type <type> [mesh_id <meshid>]",
+COMMAND(interface, add, "<name> type <type> [mesh_id <meshid>] [flags ...]",
 	NL80211_CMD_NEW_INTERFACE, 0, CIB_NETDEV, handle_interface_add);
 
-static int handle_interface_del(struct nl_cb *cb,
+static int handle_interface_del(struct nl80211_state *state,
+				struct nl_cb *cb,
 				struct nl_msg *msg,
 				int argc, char **argv)
 {
@@ -143,7 +212,8 @@ static int print_iface_handler(struct nl_msg *msg, void *arg)
 	return NL_SKIP;
 }
 
-static int handle_interface_info(struct nl_cb *cb,
+static int handle_interface_info(struct nl80211_state *state,
+				 struct nl_cb *cb,
 				 struct nl_msg *msg,
 				 int argc, char **argv)
 {
@@ -152,58 +222,36 @@ static int handle_interface_info(struct nl_cb *cb,
 }
 TOPLEVEL(info, NULL, NL80211_CMD_GET_INTERFACE, 0, CIB_NETDEV, handle_interface_info);
 
-static int handle_interface_set(struct nl_cb *cb,
+static int handle_interface_set(struct nl80211_state *state,
+				struct nl_cb *cb,
 				struct nl_msg *msg,
 				int argc, char **argv)
 {
-	enum nl80211_mntr_flags flag;
-	struct nl_msg *flags;
-	int err;
-
 	if (!argc)
 		return 1;
 
-	flags = nlmsg_alloc();
-	if (!flags) {
-		fprintf(stderr, "failed to allocate flags\n");
-		return 2;
-	}
-
 	NLA_PUT_U32(msg, NL80211_ATTR_IFTYPE, NL80211_IFTYPE_MONITOR);
 
-	while (argc) {
-		int ok = 0;
-		for (flag = __NL80211_MNTR_FLAG_INVALID + 1;
-		     flag < NL80211_MNTR_FLAG_MAX; flag++) {
-			if (strcmp(*argv, mntr_flags[flag]) == 0) {
-				ok = 1;
-				NLA_PUT_FLAG(flags, flag);
-				break;
-			}
-		}
-		if (!ok) {
-			fprintf(stderr, "unknown flag %s\n", *argv);
-			err = 2;
-			goto out;
-		}
-		argc--;
-		argv++;
+	switch (parse_mntr_flags(&argc, &argv, msg)) {
+	case 0:
+		return 0;
+	case -ENOMEM:
+		fprintf(stderr, "failed to allocate flags\n");
+		return 2;
+	case -EINVAL:
+		fprintf(stderr, "unknown flag %s\n", *argv);
+		return 2;
+	default:
+		return 2;
 	}
-
-	nla_put_nested(msg, NL80211_ATTR_MNTR_FLAGS, flags);
-
-	err = 0;
-	goto out;
  nla_put_failure:
-	err = -ENOBUFS;
- out:
-	nlmsg_free(flags);
-	return err;
+	return -ENOBUFS;
 }
 COMMAND(set, monitor, "<flag> [...]",
 	NL80211_CMD_SET_INTERFACE, 0, CIB_NETDEV, handle_interface_set);
 
-static int handle_interface_meshid(struct nl_cb *cb,
+static int handle_interface_meshid(struct nl80211_state *state,
+				   struct nl_cb *cb,
 				   struct nl_msg *msg,
 				   int argc, char **argv)
 {
